#include "dynamic_lib.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "assert.h"

#define UNUSED(a) ((void) (a))

static char *alloc_string(size_t length)
{
    return (char *) malloc(sizeof(char) * (length + 1));
}

char *dyn_strcpy(const char *str)
{
    assert(str != NULL);

    size_t size = strlen(str);

    char *buffer = alloc_string(size);
    if (buffer == NULL) {
        perror("Memory allocation failed!");
        return NULL;
    }

    strncpy(buffer, str, size);
    buffer[size] = '\0';

    return buffer;
}

char *dyn_strjoin(const char *pre, const char *post)
{
    assert(pre != NULL);
    assert(post != NULL);

    char *buffer = alloc_string(strlen(pre) + strlen(post));
    if (buffer == NULL) {
        perror("Memory allocation failed!");
        return NULL;
    }

    strcpy(buffer, pre);
    strcat(buffer, post);
    return buffer;
}

char *read_line(void)
{
    char *buffer = NULL;
    size_t alloc_size = 1;
    size_t size = 0;
    int ch = 0;

    while ((ch = getchar()) != EOF && ch != '\n') {
        if ((size + 1) >= alloc_size) {
            alloc_size *= 2;
            char *tmp_buffer = (char *) realloc(buffer, (alloc_size + 1) * sizeof(char));
            if (tmp_buffer == NULL) {
                perror("Memory allocation failed!");
                free(buffer);
                return NULL;
            }
	        buffer = tmp_buffer;
        }

        buffer[size++] = (char) ch;
    }

    if (buffer == NULL) {
        return dyn_strcpy(""); // First character was new line or EOF
    }

    buffer[size] = '\0';
    return buffer;
}

void **dyn_alloc2d(size_t rows, const size_t row_sizes[rows])
{
    assert(rows != 0);
    assert(row_sizes != NULL);

    void **memory = (void **) calloc(rows, sizeof(void *));
    if (memory == NULL) {
        perror("Memory allocation failed!");
        return NULL;
    }

    size_t i = 0;
    for (; i < rows; i++) {
        size_t row_size = row_sizes[i];
        if (row_size == 0) {
            continue;
        }
        memory[i] = (void *) malloc(row_size);
        if (memory[i] == NULL) {
            perror("Memory allocation failed!");
            break;
        }
    }

    if (i != rows) {
        dyn_free2d(memory, i);
        if (i == 0) {
            free(memory);
        }

        memory = NULL;
    }

    return memory;
}

int dyn_free2d(void **memory, size_t rows)
{
    if (memory == NULL || rows == 0) {
        return 1;
    }

    for (size_t row = 0; row < rows; row++) {
        free(memory[row]);
    }

    free(memory);
    return 0;
}

/**
 * BONUSES
 */

unsigned **pascal_triangle(size_t depth)
{
    assert(depth != 0);

    size_t *sizes = malloc(depth * sizeof(size_t));
    if (sizes == NULL) {
        perror("Memory allocation failed!");
        return NULL;
    }
    for (size_t i = 0; i < depth; i++) {
        sizes[i] = (i + 1) * sizeof(unsigned);
    }

    unsigned **triangle = (unsigned **) dyn_alloc2d(depth, sizes);
    if (triangle == NULL) {
        return NULL;
    }

    for (size_t i = 0; i < depth; i++) {
        for (size_t j = 0; j <= i; j++) {
            triangle[i][j] = (j == 0 || i == j) ? 1 : triangle[i - 1][j - 1] + triangle[i - 1][j];
        }
    }
    return triangle;
}

size_t string_count_substr(const char *original, const char *substring)
{
    assert(original != NULL);
    assert(substring != NULL);

    size_t len = strlen(substring);
    if (len == 0) {
        return 0;
    }

    size_t counter = 0;
    while (*original != '\0') {
        if (strncmp(original, substring, len) == 0) {
            counter++;
        }
        original++;
    }

    return counter;
}

char **string_split(const char *orig, const char *splitter, size_t *size)
{
    assert(orig != NULL);
    assert(splitter != NULL);
    assert(size != NULL);

    size_t split_length = strlen(splitter);
    char **data = NULL;
    size_t alloc_length = 64;
    size_t actual_length = 0;
    /* How many there will be substrings: +1 for the last substring and +1 for NULL terminator => +2 */
    size_t subs = string_count_substr(orig, splitter) + 2;
    data = (char **) malloc(sizeof(char *) * subs);
    if (data == NULL) {
        perror("Memory allocation failed!");
        return NULL;
    }

    *size = 0;
    data[*size] = alloc_string(alloc_length); // alocate some default amount of the memory
    if (data[*size] == NULL) {
        perror("Memory allocation failed!");

        free(data);
        return NULL;
    }
    data[*size][0] = '\0';

    while (*orig != '\0') {
        // compare whether I found the splitter
        if (strncmp(orig, splitter, split_length) == 0) { // I found the splitter
            orig += split_length;                         // move in original string by splitter length
            (*size)++;                                    // increase the number of substrings
            actual_length = 0;                            // reset default values
            alloc_length = 64;
            data[*size] = alloc_string(alloc_length);
            if (data[*size] == NULL) {
                perror("Memory allocation failed!");

                dyn_free2d((void **) data, *size);
                return NULL;
            }
            data[*size][0] = '\0';

            continue;
        }

        // realocate if needed
        if (actual_length >= (alloc_length - 2)) {
            alloc_length *= 2;
            char *new_string = realloc(data[*size], sizeof(char) * alloc_length);
            if (new_string == NULL) {
                perror("Memory reallocation failed!");

                dyn_free2d((void **) data, *size + 1);
                return NULL;
            }

            data[*size] = new_string;
        }

        data[*size][actual_length++] = *orig;
        orig++;
        data[*size][actual_length] = 0;
    }

    (*size)++;
    data[*size] = NULL;

    return data;
}

char *string_replace_all(const char *original, const char *from, const char *to)
{
    assert(original != NULL);
    assert(from != NULL);
    assert(to != NULL);

    size_t parts_count = 0;
    char **parts = string_split(original, from, &parts_count);
    if (parts == NULL) {
        return NULL;
    }

    // substring was not found
    if (parts_count == 1) {
        char *copy_of_original = parts[0];

        free(parts);
        return copy_of_original;
    }

    // prefix all parts with the new substring
    for (size_t i = 1; i < parts_count; i++) {
        char *old_part = parts[i];
        parts[i] = dyn_strjoin(to, old_part);
        free(old_part);

        if (parts[i] == NULL) {
            dyn_free2d((void **) parts, parts_count);
            return NULL;
        }
    }

    char *built_prefix = dyn_strcpy(parts[0]);
    if (built_prefix == NULL) {
        dyn_free2d((void **) parts, parts_count);
        return NULL;
    }

    for (size_t i = 1; i < parts_count; i++) {
        char *last_prefix = built_prefix;
        built_prefix = dyn_strjoin(last_prefix, parts[i]);
        free(last_prefix);

        if (built_prefix == NULL) {
            dyn_free2d((void **) parts, parts_count);
            return NULL;
        }
    }
    dyn_free2d((void **) parts, parts_count);

    return built_prefix;
}
